```@meta
CurrentModule = Parquet2
```

# Parquet2

This package implements the [parquet tabular data storage
format](https://github.com/apache/parquet-format) in pure Julia.  A particular emphasis is placed on
selective loading, schema introspection and the flexibility to load data from a variety of sources such
as S3 or min.io.  The package is based on a lazy data loading and caching system that allows users
to load only the needed subsets of data rather than entire files.  This is particularly important
for the parquet format since it is often used to store "large" quantities of data (i.e. much larger
than is likely to fit in memory).

Parquet data loaded by Parquet2.jl is organized analogously to the underlying binary data schema:
```
Dataset (← Tables.jl table)
    ⎸ RowGroup (← Tables.jl table)
    ⎸   ⎸ Column1 (← lazy AbstractVector)
    ⎸   ⎸ Column2 (← lazy AbstractVector)
    ⎸   ⎸ ⋮
    ⎸   ⎸ ColumnN (← lazy AbstractVector)
    ⎸ RowGroup2 (← Tables.jl table)
    ⎸ ⋮
    ⎸ RowGroupN (← Tables.jl table)
    ⎸   ⋮
```
- [`Dataset`](@ref): a [Tables.jl table](https://tables.juliadata.org/dev/) and an indexable set of
    `RowGroup`.
- [`RowGroup`](@ref): a Tables.jl table and an indexable (by string name or integer) set of
    `Column`.
- [`Column`](@ref): an abstraction for getting `AbstractVector` objects, which contain the
    deserialized contents of the parquet column.

Parquet files are divided into ``n \ge 0`` `RowGroup`s each of which represents a contiguous
subset of rows in the full table.  `RowGroup`s typically correspond to files.  Parquet2.jl
treats the `RowGroup`s as full tables in-and-of themselves and they can be loaded just like the full
dataset.

## Initialing a Dataset
The first step in reading parquet data is creating a `Dataset` object which contains only metadata.
```julia
using FilePathsBase
using Parquet2: Dataset

ds = Dataset("/path/to/file")  # this only loads metadata
ds = Dataset(p"s3://path/to/file")  # can understand path types (may require extension modules)
ds = Dataset("/path/to/dir.parquet/")  # parquets may be a directory that contain top-level metdata
```

Showing the dataset will display the schema, here is one of the validation datasets:
```julia
◖◗ ds = Dataset("test/data/std_fastparquet.parq")
≔ Dataset (31016 bytes)
	1. "floats": Union{Missing, Float64}
	2. "floats_missing": Union{Missing, Float64}
	3. "ints": Union{Missing, Int64}
	4. "ints_missing": Union{Missing, Float64}
	5. "strings": Union{Missing, String}
	6. "strings_missing": Union{Missing, String}
	7. "timestamps": Union{Missing, DateTime}
	8. "timestamps_missing": Union{Missing, DateTime}
	9. "missings": Missing
	10. "dictionary": Union{Missing, Int64}
	11. "dictionary_missing": Union{Missing, Int64}
```

The dataset is an indexable, iterable collection of `RowGroup` objects
```julia
length(ds)  # gives the number of row groups

rg = ds[1]  # first row group


◖◗ for rg ∈ ds
        println(rg)
    end
≔ RowGroup (14486 bytes) (250 rows)
	1. "floats": Union{Missing, Float64}
	2. "floats_missing": Union{Missing, Float64}
	3. "ints": Union{Missing, Int64}
	4. "ints_missing": Union{Missing, Float64}
	5. "strings": Union{Missing, String}
	6. "strings_missing": Union{Missing, String}
	7. "timestamps": Union{Missing, DateTime}
	8. "timestamps_missing": Union{Missing, DateTime}
	9. "missings": Missing
	10. "dictionary": Union{Missing, Int64}
	11. "dictionary_missing": Union{Missing, Int64}

≔ RowGroup (14480 bytes) (250 rows)
	1. "floats": Union{Missing, Float64}
	2. "floats_missing": Union{Missing, Float64}
	3. "ints": Union{Missing, Int64}
	4. "ints_missing": Union{Missing, Float64}
	5. "strings": Union{Missing, String}
	6. "strings_missing": Union{Missing, String}
	7. "timestamps": Union{Missing, DateTime}
	8. "timestamps_missing": Union{Missing, DateTime}
	9. "missings": Missing
	10. "dictionary": Union{Missing, Int64}
	11. "dictionary_missing": Union{Missing, Int64}
```

Note that each `RowGroup` matches the schema of the `Dataset`.  The `RowGroup`s belonging to a
dataset need not have the same number of rows.


## Accessing Data
[`Dataset`](@ref) and [`RowGroup`](@ref) both satisfy the
[Tables.jl](https://github.com/JuliaData/Tables.jl) columnar table interface and can therefore be
easily converted to tables.
```julia
using Tables, DataFrames

using Parquet2: Dataset

# Dataset is an abstraction to provide an interface for loading data
ds = Dataset("/path/to/file")

sch = Tables.schema(ds)  # get the Tables.jl Schema object

c = Tables.getcolumn(ds, :col1)  # load *only* col1; others are not touched

c = Parquet2.load(ds, "col1")  # equivalent to the above


df = DataFrame(ds; copycols=false)  # load the entire table as a DataFrame

# it is suggested to do `copycols=false` unless you intend to write to the DataFrame
df = DataFrame(ds)

# columns are loaded in parallel by default, but this can be disabled
ds = Dataset("/path/to/file"; parallel_column_loading=false)

df1 = DataFrame(ds[1])  # load the first RowGroup as a DataFrame

meta = Parquet2.metadata(ds)  # auxiliary metadata can be accessed thusly
colmeta = Parquet2.metadata(ds[1]["col1"])  # auxiliary metadata by column
```

!!! note

    Columns will be read from the file or buffer and allocated every time `Parquet2.load` or,
    equivalently `Tables.getcolumn` is called on a `Dataset` or `RowGroup`.  These operations should
    be considered expensive, and the API assumes they will be used sparingly.  If you want to call
    something that frequently accesses columns, you should first make all `Parquet2.load` calls,
    returning fully materialized columns and then call your oprations on these.  For example, when
    performing operations on columns, it is usually better to first collect the table in-memory with
    `Tables.columntable` or `DataFrame`.


### Using TableOperations.jl to load Specific Columns
This package supports loading of specific columns leaving the rest untouched.  With the exception of
the top-level metadata, only data from selected columns is loaded.

While columns can easily be loaded individually with `Tables.getcolumn` or [`Parquet2.load`](@ref),
it is more convenient to select a subset of them using
[TableOperations.jl](https://github.com/JuliaData/TableOperations.jl).  In particular, see
[`TableOperations.select`](https://github.com/JuliaData/TableOperations.jl#tableoperationsselect).

For example
```julia
using TableOperations; const TO = TableOperations

ds = Parquet2.Dataset("/path/to/file")

# load *only* columns 1 and 2, others are skipped entirely
df = ds |> TO.select(:col1, :col2) |> DataFrame

# TableOperations is a dependency of Parquet2, so you can also do
df = ds |> Parquet2.select(:col3, :col4) |> DataFrame
```

## Writing Data
Files can be written with [`Parquet2.writefile`](@ref).  Tables being written must satisfy [the
Tables.jl interface](https://tables.juliadata.org/stable/) (most Julia-ecosystem tables such as
`DataFrame`s already do this).  They can be written to any filesystem which can be written to with
the `FilePatshBase` interface.

For example
```julia
using Parquet2, DataFrames

using Parquet2: writefile

tbl = DataFrame(A=1:5, B=rand(5))
writefile("testfile.parq", tbl)  # write to disk at testfile.parq

io = IOBuffer()
writefile(io, tbl)  # write to IO buffer

writefile(Vector{UInt8}, tbl)  # write to an array

tbl = (A=1:5, B=rand(5))  # a NamedTuple is the archetyple Tables.jl copmatible object

writefile("testfile2.parq", tbl;
          npages=2,  # number of pages per column
          compression_codec=Dict("A"=>:zstd, "B"=>:snappy),  # compression codec per column
          column_metadata="A"=>Dict("frank"=>"reynolds"),  # per column auxiliary metadata
          metadata=Dict("charlie"=>"kelley"),  # file wide metadata
         )
```

Note that any option which can be defined per-column such as `npages` and `compression_codec` can be
specified either by value, in which case it'll be applied to all columns, dictionary per column, or
pair for a single column.

### Writing Row Groups
`writefile` uses the Tables.jl `Tables.partitions` function to decide how to group rows into
`RowGroup`s.  Each partition is written as a single row group.

## Column Statistics
The parquet format includes the ability to store computed statistics values in metadata so that they
can be retrieved without the need to re-compute them.  Statistics which can be stored are
- Minimum value.
- Maximum value.
- Number of missing values.
- Number of distinct values.

This package provides the option to compute these values when data is serialized, as well as the
option to wrap the output `AbstractVector`s in a `struct` which will allow these to be retrieved
rather than being re-computed with methods from `Base`.

```julia
tbl = (A=1:5, B=6:10, C=1.0:5.0)

# compute statistics when writing only for columns A and C
v = writefile(Vector{UInt8}, tbl; compute_statistics=["A", "C"])

# only wrap column A
# (note that if `copycols=true` the wrapper will get destroyed)
df = DataFrame(Dataset(v; use_statistics=["A"]); copycols=false)

minimum(df.A)  # this is simply retrieved rather than being re-computed
minimum(df.B)  # this was not stored and had to be re-computed
maximum(df.C)  # this was computed, but we didn't include it so it's computed again here

stats = Parquet2.ColumnStatistics(df.C)  # we can still access that stats though
minimum(stats)
maximum(stats)
count(ismissing, stats)
Parquet2.ndistinct(stats)
```

## Extensions
The following packages are available for loading data from alternative (i.e. not the local
file-system) sources:
- [ParquetS3.jl](https://gitlab.com/ExpandingMan/ParquetS3.jl)
