
"""
    Dataset <: ParquetTable

A complete parquet dataset created from top-level parquet metadata.  Each `Dataset` is an indexable collection of
[`RowGroup`](@ref)s each of which is a Tables.jl compatible columnar table in its own right.  The
`Dataset` is a [Tables.jl compatible](https://tables.juliadata.org/dev/) columnar table consisting of (lazily
by default) concatenated `RowGroup`s.  A `Dataset` can consist of data in any number of files depending on the
directory structure of the referenced parquet.

## Constructors
```julia
Dataset(fm::FileManager; kw...)
Dataset(p::AbstractPath; kw...)
Dataset(v::AbstractVector{UInt8}; kw...)
Dataset(io::IO; kw...)
Dataset(str::AbstractString; kw...)
```

## Arguments
- `fm`: A [`FileManager`](@ref) object describing a set of files to be loaded.
- `p`: Path to main metadata file or directory containing a `_metadata` file.  Loading behavior will depend on
    the type of path provided.
- `v`: An in-memory (or memory mapped) byte buffer.
- `io`: An `IO` object from which data can be loaded.
- `str`: File or directory path as a string.  Converted to `AbstractPath` with `Path(str)`.

### Keyword Arguments
The following keyword arguments are applicable for the dataset as a whole:
- `support_legacy` (`true`): Some parquet writers take bizarre liberties with the metadata, in particular
    many JVM-based writers use a specialized `UInt96` encoding of timestamps even though this is not described
    by the metadata.  When this option is `false` the metadata will be interpreted strictly.
- `use_mmap` (`true`): Whether to use memory mapping for reading the file.  Only applicable for files on the
    local file system.  In some cases enabling this can drastically increase read performance.
- `subset_length` (`nothing`): The size (in bytes) of file subsets which are fetched from the data source.
    If `nothing` an appropriately large value will be automatically chosen.
- `max_subsets` (`typemax(Int)`): The maximum number of buffer subsets which can be simultaneously loaded.
- `fetch_by_column` (`false`): If `true`, the reader will attempt to access only the data for requested columns.
    Can be helpful in IO-limited cases when reading few columns from large files with many columns.
- `preload` (`false`):  Whether all data should be fetched on constructing the `Dataset` regardless of the
    above options.
- `load_initial` (`nothing`):  Whether the `RowGroup`s should be eagerly loaded into memory.  If `nothing`, this
    will be done only for parquets consisting of a single file.
- `parallel_column_loading`: Whether columns should be loaded using thread-based parallelism.  If `nothing`, this
    is true as long as Julia has multiple threads available to it.

The following keyword arguments are applicable to specific columns.  These can be passed either as a single value,
a `NamedTuple`, `AbstractDict` or `ColumnOption`.  See [`ColumnOption`](@ref) for details.
- `allow_string_copying` (`false`): Whether strings will be copied.  If `false` a reference to the underlying data
    buffer needs to be maintained, meaning it can't be garbage collected to free up memory.  Note also that
    there will potentially be a large number of references stored in the output colun if this is `false`,
    so setting this to `true` reduces garbage collector overhead.
- `lazy_dictionary` (`true`):  Whether output columns will use a Julia categorical array representation which in
    some cases can ellide a large number of allocations.
- `parallel_page_loading` (`false`): Whether data pages in the column should be loaded in parallel.  This comes
    with some additional overhead including an extra iteration over the entire page buffer, so it is of dubious
    benefit to turn this on, but it may be helpful in cases in which there is a large number of pages.
- `use_statistics` (`false`): Whether statistics included in the metadata will be used in the loaded column
    `AbstractVector`s so that statistics can be efficiently retrieved rather than being re-computed.
    Note that if this is `true` this will only be done for columns for which statistics are available.
    Otherwise, statistics can be retrieved with `ColumnStatistics(col)`.
- `eager_page_scanning` (`true`): It is not in general possible to infer all page metadata without iterating over
    the columns entire data buffer.  This can be elided, but doing so limits what can be done to accommodate data
    loaded from the column.  Turning this option off will reduce the overhead of loading metadata for the column
    but may increase the cost of allocating the output.  If `false` specialized string and dictionary outputs
    will not be used (loading the column will be maximally allocating).

## Usage
```julia
ds = Dataset("/path/to/parquet")
ds = Dataset(p"s3://path/to/parquet")  # understands different path types

length(ds)  # gives number of row groups

rg = ds[1]  # index to get row groups

for rg ∈ ds  # is an indexable, iterable collection of row groups
    println(rg)
end

df = DataFrame(ds)  # Tables.jl compatible, is concatenation of all row groups

# use TableOperations.jl to load only selected columns
df = ds |> TableOperations.select(:col1, :col2) |> DataFrame
```
"""
struct Dataset{ℱ<:FileManager} <: ParquetTable
    file_manager::ℱ
    meta_orig::Meta.FileMetaData
    schema::SchemaNode
    row_groups::Vector{RowGroup}
    row_group_index::Dict{AbstractPath,Set{Int}}
    name_index::NameIndex
    partition_tree::PartitionNode
    partition_column_names::OrderedSet{String}
    metadata::Dict{String,Any}
end

function Dataset(fm::FileManager)
    v = get(fm)
    m = readmeta(v; check=true)
    opts = ReadOptions(fm)
    r = SchemaNode(m.schema; support_legacy=opts.support_legacy)
    ptree = PartitionNode(fm)
    if _should_load_initial(fm)
        rgs = m.row_groups |> Map() do rg
            RowGroup(fm, r, rg, ptree; parallel_column_loading=opts.parallel_column_loading)
        end |> collect
        ridx = Dict{AbstractPath,Set{Int}}(fm.main_path=>Set(1:length(rgs)))
    else
        rgs = Vector{RowGroup}()
        ridx = Dict{AbstractPath,Set{Int}}()
    end
    pnames = OrderedSet(columnnames(ptree))
    nidx = (pnames, children(r) |> Map(name)) |> Cat() |> collect |> NameIndex
    Dataset{typeof(fm)}(fm, m, r, rgs, ridx, nidx, ptree, pnames, unpack_thrift_metadata(m))
end
function Dataset(p::AbstractPath; kw...)
    opts = filterkw(ReadOptions, kw)
    Dataset(FileManager(p, opts))
end
Dataset(io::IO; kw...) = Dataset(FileManager(read(io); kw...))
Dataset(v::AbstractVector{UInt8}; kw...)  = Dataset(FileManager(v; kw...))
Dataset(p::AbstractString; kw...) = Dataset(AbstractPath(p); kw...)

ReadOptions(ds::Dataset) = ReadOptions(ds.file_manager)

FileManager(ds::Dataset) = ds.file_manager

filelist(ds::Dataset) = filelist(FileManager(ds))

# not sure this makes much sense as an API but it sure is convenient
Base.get(ds::Dataset) = get(ds.file_manager)
Base.get(ds::Dataset, k) = get(ds.file_manager, k)

function _update_row_group_index!(ds::Dataset, p::AbstractPath, n::Integer)
    n == 0 && return nothing
    m = length(ds.row_groups)
    s = get(ds.row_group_index, p, Set{Int}())
    t = Set((m+1):(m+n))
    if isempty(s)
        ds.row_group_index[p] = t
    else
        union!(s, t)
    end
    nothing
end

"""
    append!(ds::Parquet2.Dataset, p; check=true)

Append all row groups from the file `p` to the dataset row group metadata.  If `check`, will check if path is a valid
parquet file first.  `p` must be a path that was discovered during the initial construction of the dataset.
"""
function Base.append!(ds::Dataset, p::AbstractPath=ds.file_manager.main_path; check::Bool=true)
    p ∈ keys(ds.row_group_index) && throw(ArgumentError("row groups for \"$p\" already exist in dataset"))
    v = get(ds.file_manager, p)
    m = readmeta(v; check)
    𝒻 = rg -> RowGroup(ds.file_manager, ds.schema, rg, ds.partition_tree; current_file=p)
    rgs = m.row_groups |> Map(𝒻) |> collect
    _update_row_group_index!(ds, p, length(rgs))
    append!(ds.row_groups, rgs)
    ds
end
Base.append!(ds::Dataset, p::AbstractString; kw...) = append!(ds, Path(p); kw...)
function Base.append!(ds::Dataset, sp::Pair{<:AbstractString,<:AbstractString})
    foreach(f -> append!(ds, f), filelist(ds) |> Filter(s -> occursin(directorystring(sp), string(s))))
end

"""
    dirname(ds::Dataset)

Get the parent directory of the dataset.
"""
Base.dirname(ds::Dataset) = dirname(ds.file_manager)

partition_column_names(ds::Dataset) = ds.partition_column_names

"""
    pathof(ds::Parquet2.Dataset)

Get the path of the file containing the main metadata for dataset `ds`.
"""
Base.pathof(ds::Dataset) = isnothing(ds.path) ? nothing : ds.path

nbytes(ds::Dataset) = ds |> FileManager |> get |> length

_min_parquet_size() = 2*length(MAGIC) + FOOTER_LENGTH

function checkparquet(v::CacheVector)
    isempty(v) && throw(ArgumentError("invalid parquet: provided buffer is empty"))
    if length(v) < _min_parquet_size()
        throw(ArgumentError("invalid parquet: buffer of length $(length(v)) is shorter than minimum "*
                            "size $(_min_parquet_size())"))
    end
    mgc = v[(end-length(MAGIC)+1):end]
    if mgc ≠ MAGIC
        throw(ArgumentError("invalid parquet: final bytes are \"$(String(mgc))\", expect \"PAR1\""))
    end
end

function readmetalength(v::CacheVector)
    j = length(v) - length(MAGIC) - FOOTER_LENGTH + 1
    only(reinterpret(Int32, v[j:(j+3)]))
end

function readmeta(v::CacheVector; check::Bool=true)
    check && checkparquet(v)
    ℓ = readmetalength(v)
    a = length(v) - length(MAGIC) - FOOTER_LENGTH - ℓ
    m = readthrift(CacheIO(v, a), Meta.FileMetaData)
    @debug("read parquet Dataset metadata")
    m
end

"""
    metadata(ds::Dataset)

Get the auxiliary key-value metadata for the dataset.
"""
metadata(ds::Dataset) = ds.metadata

"""
    metadata(ds::Dataset, k::AbstractString, default=nothing)

Get the key `k` from the key-value metadata for the dataset, returning `default` if not present
"""
metadata(ds::Dataset, k::AbstractString, default=nothing) = get(metadata(ds), k, default)

rowgroups(ds::Dataset) = ds.row_groups

nrowgroups(ds::Dataset) = length(ds.row_groups)

DataAPI.nrow(ds::Dataset) = sum(nrow, rowgroups(ds))

RowGroup(ds::Dataset, n::Integer) = rowgroups(ds)[n]

Base.length(ds::Dataset) = nrowgroups(ds)

Base.getindex(ds::Dataset, n::Integer) = RowGroup(ds, n)

Base.close(ds::Dataset) = close(ds.file_manager)

function Base.iterate(ds::Dataset, n::Integer=1)
    n > nrowgroups(ds) && return nothing
    RowGroup(ds, n), n+1
end

Column(ds::Dataset, r::Integer, c::Union{Integer,AbstractString}) = Column(RowGroup(ds, r), c)

function PageLoader(ds::Dataset, r::Integer, c::Union{Integer,AbstractString}, p::Integer=1)
    PageLoader(ds.schema, Column(ds, r, c), p)
end

"""
    load(ds::Dataset, n)

Load the (complete, all `RowGroup`s) column `n` (integer or string) from the dataset.
"""
function load(ds::Dataset, n::Union{Integer,AbstractString})
    if nrowgroups(ds) == 0
        Vector{juliamissingtype(ds, n)}()
    elseif nrowgroups(ds) == 1
        load(RowGroup(ds, 1), n)
    else
        ChainedVector(map(rg -> load(rg, n), rowgroups(ds)))
    end
end

Tables.partitions(ds::Dataset) = rowgroups(ds)

_useparallel(opt::Union{Nothing,Bool}) = isnothing(opt) ? (Base.Threads.nthreads() > 1) : opt

useparallel(ds::Dataset) = _useparallel(ReadOptions(ds).parallel_column_loading)

"""
    readfile(filename; kw...)
    readfile(io::IO; kw...)

An alias for [`Dataset`](@ref).  All arguments are the same, so see those docs.

This function is provided for consistency with the [`writefile`](@ref) function.
"""
readfile(a...; kw...) = Dataset(a...; kw...)
